@@ -658,6 +658,11 @@ pub unsafe extern "C" fn slint_interpreter_model_notify_row_added(
     notify.as_model_notify().row_added(row, count);
 }
 
+#[no_mangle]
+pub unsafe extern "C" fn slint_interpreter_model_notify_reset(notify: &ModelNotifyOpaque) {
+    notify.as_model_notify().reset();
+}
+
 #[no_mangle]
 pub unsafe extern "C" fn slint_interpreter_model_notify_row_removed(
     notify: &ModelNotifyOpaque,
